# cython: language_level=3
import numpy as np
cimport numpy as np
cimport cython
ctypedef np.float_t FLOAT_t
ctypedef np.int64_t INT_t
from scipy import optimize
from scipy.ndimage.filters import gaussian_filter1d
from scipy.signal import convolve, kaiser

cdef int within_bounds(np.ndarray[FLOAT_t, ndim=1] res, np.ndarray[FLOAT_t, ndim=2] bnds):
    for i,j in zip(res, bnds):
        if j[0] is not None and i < j[0]:
            return 0
        if j[1] is not None and i > j[1]:
            return 0
    return 1

cpdef np.ndarray[FLOAT_t, ndim=1] gauss(np.ndarray[FLOAT_t, ndim=1] x, FLOAT_t amp, FLOAT_t mu, FLOAT_t std):
    cdef np.ndarray[FLOAT_t, ndim=1] y = amp*np.exp(-(x - mu)**2/(2*std**2))
    return y

cdef np.ndarray[FLOAT_t, ndim=1] adjust_baseline(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] mus, FLOAT_t slope, FLOAT_t intercept, INT_t index):#np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] mus, float slope, float intercept, int index):
    cdef np.ndarray[FLOAT_t, ndim=1] y
    cdef INT_t left, right

    if index == len(mus)-1:
        right = -1
    else:
        right = np.searchsorted(xdata, (mus[index+1]+mus[index])/2.0, side='right')
    if index == 0:
        left = 0
    else:
        left = np.searchsorted(xdata, (mus[index-1]+mus[index])/2.0, side='right')
    y = slope*xdata+intercept
    y[:left] = 0
    if right != -1:
        y[right:] = 0
    return y

cpdef np.ndarray[FLOAT_t, ndim=1] gauss_ndim(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] params):
    cdef np.ndarray[FLOAT_t, ndim=1] amps
    cdef np.ndarray[FLOAT_t, ndim=1] mus
    cdef np.ndarray[FLOAT_t, ndim=1] sigmas
    cdef FLOAT_t amp, mu, sigma
    cdef np.ndarray[FLOAT_t, ndim=1] data = np.zeros(len(xdata))

    amps, mus, sigmas = params[::3], params[1::3], params[2::3]
    for amp, mu, sigma in zip(amps, mus, sigmas):
        data += gauss(xdata, amp, mu, sigma)
    return data

cpdef np.ndarray[FLOAT_t, ndim=1] gauss_bl_ndim(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] params):
    cdef np.ndarray[FLOAT_t, ndim=1] amps
    cdef np.ndarray[FLOAT_t, ndim=1] mus
    cdef np.ndarray[FLOAT_t, ndim=1] sigmas
    cdef np.ndarray[FLOAT_t, ndim=1] slopes
    cdef np.ndarray[FLOAT_t, ndim=1] intercepts
    cdef np.ndarray[INT_t, ndim=1] sort_order
    cdef FLOAT_t amp, mu, sigma
    cdef np.ndarray[FLOAT_t, ndim=1] data = np.zeros(len(xdata))

    amps, mus, sigmas, slopes, intercepts = params[::5], params[1::5], params[2::5], params[3::5], params[4::5]
    sort_order = np.argsort(mus)

    for index in sort_order:
        amp, mu, sigma = amps[index], mus[index], sigmas[index]
        data += gauss(xdata, amp, mu, sigma)
        data += adjust_baseline(xdata, mus, slopes[index], intercepts[index], index)
    return data

cpdef FLOAT_t gauss_func(np.ndarray[FLOAT_t, ndim=1] guess, np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] ydata, int baseline_correction):
    cdef np.ndarray[FLOAT_t, ndim=1] data = gauss_bl_ndim(xdata, guess) if baseline_correction else gauss_ndim(xdata, guess)
    # absolute deviation as our distance metric. Empirically found to give better results than
    # residual sum of squares for this data.
    cdef FLOAT_t residual = sum((ydata-data)**2)
    return residual

cpdef np.ndarray[FLOAT_t, ndim=1] bigauss_jac(np.ndarray[FLOAT_t, ndim=1] params, np.ndarray[FLOAT_t, ndim=1] x, np.ndarray[FLOAT_t, ndim=1] y, baseline_correction):
    cdef np.ndarray[FLOAT_t, ndim=1] jac, common, amp_term, left_common, right_common
    cdef np.ndarray[FLOAT_t, ndim=1] lx, ly, rx, ry
    cdef float amp, mu, sigma1, sigma2
    cdef int i
    jac = np.zeros_like(params)
    common = -bigauss_ndim(x, params) + y
    for i in xrange(params.shape[0]):
        if i%4 == 0:
            amp = params[i]
        elif i%4 == 1:
            mu = params[i]
        elif i%4 == 2:
            sigma1 = params[i]
        elif i%4 == 3:
            sigma2 = params[i]
            lx = x[x<=mu]
            amp_term = np.exp(-(-mu + lx)**2/(2*sigma1**2))
            left_common = common[x<=mu]
            jac[i-3] += sum(-2*left_common*amp_term)
            jac[i-2] += sum(amp*(2*mu - 2*lx)*left_common*amp_term/sigma1**2)
            jac[i-1] += sum(-2*amp*(-mu + lx)**2*left_common*amp_term/sigma1**3)
            rx = x[x>mu]
            amp_term = np.exp(-(-mu + rx)**2/(2*sigma2**2))
            right_common = common[x>mu]
            jac[i-3] += sum(-2*right_common*amp_term)
            jac[i-2] += sum(amp*(2*mu - 2*rx)*right_common*amp_term/sigma2**2)
            jac[i] += sum(-2*amp*(-mu + rx)**2*right_common*amp_term/sigma2**3)
    return jac

cpdef np.ndarray[FLOAT_t, ndim=1] gauss_jac_old(np.ndarray[FLOAT_t, ndim=1] params, np.ndarray[FLOAT_t, ndim=1] x, np.ndarray[FLOAT_t, ndim=1] y):
    cdef np.ndarray[FLOAT_t, ndim=1] jac
    cdef float amp, mu, sigma
    jac = np.zeros_like(params)
    for i in xrange(params.shape[0]):
        if i%3 == 0:
            amp = params[i]
        elif i%3 == 1:
            mu = params[i]
        elif i%3 == 2:
            sigma = params[i]
            exp_term = np.exp(-((x-mu)**2)/(2*sigma**2))
            amp_exp_term = amp*exp_term
            jac[i-2] += sum(-2*exp_term*(y-amp_exp_term))
            jac[i-1] += sum(-2*amp*(x-mu)*exp_term*(y-amp_exp_term)/(sigma**2))
            jac[i] += sum(-2*amp*((x-mu)**2)*exp_term*(y-amp_exp_term)/(sigma**3))
    return jac

cpdef np.ndarray[FLOAT_t, ndim=1] gauss_jac(np.ndarray[FLOAT_t, ndim=1] params, np.ndarray[FLOAT_t, ndim=1] x, np.ndarray[FLOAT_t, ndim=1] y, baseline_correction):
    cdef np.ndarray[FLOAT_t, ndim=1] jac, common, amp_term
    cdef float amp, mu, sigma
    jac = np.zeros_like(params)
    common = -gauss_ndim(x, params) + y
    for i in xrange(params.shape[0]):
        if i%3 == 0:
            amp = params[i]
        elif i%3 == 1:
            mu = params[i]
        elif i%3 == 2:
            sigma = params[i]
            amp_term = np.exp(-(-mu + x)**2/(2*sigma**2))
            jac[i-2] += sum(-2*common*amp_term)
            jac[i-1] += sum(amp*(2*mu - 2*x)*common*amp_term/sigma**2)
            jac[i] += sum(-2*amp*(-mu + x)**2*common*amp_term/sigma**3)
    return jac

cdef np.ndarray[FLOAT_t, ndim=1] bigauss(np.ndarray[FLOAT_t, ndim=1] x, FLOAT_t amp, FLOAT_t mu, FLOAT_t stdl, FLOAT_t stdr):
    cdef FLOAT_t sigma1 = stdl
    cdef FLOAT_t sigma2 = stdr
    cdef np.ndarray[FLOAT_t, ndim=1] lx = x[x<=mu]
    cdef np.ndarray[FLOAT_t, ndim=1] left = amp*np.exp(-(lx-mu)**2/(2*sigma1**2))
    cdef np.ndarray[FLOAT_t, ndim=1] rx = x[x>mu]
    cdef np.ndarray[FLOAT_t, ndim=1] right = amp*np.exp(-(rx-mu)**2/(2*sigma2**2))
    cdef np.ndarray[FLOAT_t, ndim=1] y = np.concatenate([left, right], axis=0)
    return y

cpdef np.ndarray[FLOAT_t, ndim=1] bigauss_ndim(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] params):
    cdef np.ndarray[FLOAT_t, ndim=1] amps
    cdef np.ndarray[FLOAT_t, ndim=1] mus
    cdef np.ndarray[FLOAT_t, ndim=1] sigmasl
    cdef np.ndarray[FLOAT_t, ndim=1] sigmasr
    amps, mus, sigmasl, sigmasr = params[::4], params[1::4], params[2::4], params[3::4]
    cdef np.ndarray[FLOAT_t, ndim=1] data = np.zeros(len(xdata))
    cdef np.ndarray[FLOAT_t, ndim=1] fit = np.zeros(len(xdata))
    for amp, mu, sigma1, sigma2 in zip(amps, mus, sigmasl, sigmasr):
        fit = bigauss(xdata, amp, mu, sigma1, sigma2)
        if fit.shape[0] == data.shape[0]:
            data += bigauss(xdata, amp, mu, sigma1, sigma2)
    return data

cpdef np.ndarray[FLOAT_t, ndim=1] bigauss_bl_ndim(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] params):
    cdef np.ndarray[FLOAT_t, ndim=1] amps
    cdef np.ndarray[FLOAT_t, ndim=1] mus
    cdef np.ndarray[FLOAT_t, ndim=1] sigmasl
    cdef np.ndarray[FLOAT_t, ndim=1] sigmasr
    cdef np.ndarray[FLOAT_t, ndim=1] slopes
    cdef np.ndarray[FLOAT_t, ndim=1] intercepts
    cdef np.ndarray[INT_t, ndim=1] sort_order
    cdef FLOAT_t amp, mu, sigma1, sigma2
    cdef np.ndarray[FLOAT_t, ndim=1] data = np.zeros(len(xdata))
    cdef np.ndarray[FLOAT_t, ndim=1] fit = np.zeros(len(xdata))
    cdef np.ndarray[FLOAT_t, ndim=1] baseline = np.zeros(len(xdata))

    amps, mus, sigmasl, sigmasr, slopes, intercepts = params[::6], params[1::6], params[2::6], params[3::6], params[4::6], params[5::6]
    sort_order = np.argsort(mus)

    for index in sort_order:
        amp, mu, sigma1, sigma2 = amps[index], mus[index], sigmasl[index], sigmasr[index]
        fit = bigauss(xdata, amp, mu, sigma1, sigma2)
        baseline = adjust_baseline(xdata, mus, slopes[index], intercepts[index], index)
        if fit.shape[0] == data.shape[0] and baseline.shape[0] == data.shape[0]:
            data += fit
            data += baseline
    return data


cpdef FLOAT_t bigauss_func(np.ndarray[FLOAT_t, ndim=1] guess, np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] ydata, int baseline_correction):
    if any([np.isnan(i) for i in guess]):
        return np.inf
    cdef np.ndarray[FLOAT_t, ndim=1] data = bigauss_bl_ndim(xdata, guess) if baseline_correction else bigauss_ndim(xdata, guess)
    cdef FLOAT_t residual = sum((ydata-data)**2)
    return residual

cpdef np.ndarray[FLOAT_t] fixedMeanFit(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] ydata,
                                       int peak_index=1, debug=False):
    cdef float rel_peak = ydata[peak_index]
    cdef float peak_loc = xdata[peak_index]
    cdef int peak_left, peak_right
    cdef float peak_min, peak_max, average, variance
    cdef np.ndarray[FLOAT_t, ndim=1] bnds

    ydata /= ydata.max()
    peak_left, peak_right = findPeak(convolve(ydata, kaiser(10, 14), mode='same'), peak_index)
    peak_min, peak_max = xdata[0], xdata[-1]
    # reset the fitting data to our bounds
    if peak_index == peak_right:
        peak_index -= peak_left-1
    else:
        peak_index -= peak_left
    xdata = xdata[peak_left:peak_right]
    ydata = ydata[peak_left:peak_right]
    if ydata.sum() == 0:
        return None
    min_spacing = min(np.diff(xdata))/2
    lb = np.fabs(peak_loc-xdata[0])
    rb = np.fabs(xdata[-1]-peak_loc)
    if lb < min_spacing:
        lb = min_spacing*5
    if rb < min_spacing:
        rb = min_spacing*5
    bnds = np.array([(rel_peak*0.75, 1.01) if rel_peak > 0 else (0.0, 1.0), (xdata[0], xdata[-1]), (min_spacing, lb), (min_spacing, rb)])
    average = np.average(xdata, weights=ydata)
    variance = np.sqrt(np.average((xdata-average)**2, weights=ydata))
    if variance == 0:
        # we have a singular peak if variance == 0, so set the variance to half of the x/y spacing
        if peak_index >= 1:
            variance = np.abs(peak_loc-xdata[peak_index-1])
        elif peak_index < len(xdata):
            variance = np.abs(xdata[peak_index+1]-peak_loc)
        else:
            # we have only 1 data point, most RT's fall into this width
            variance = 0.05
    if variance > xdata[peak_index]-peak_min or variance > peak_max-xdata[peak_index]:
        variance = xdata[peak_index]-peak_min
    cdef np.ndarray[FLOAT_t] guess = np.array([rel_peak, peak_loc, variance, variance])
    args = (xdata, ydata)
    base_opts = {'maxiter': 1000, 'ftol': 1e-20}
    routines = [('SLSQP', base_opts), ('TNC', base_opts), ('L-BFGS-B', base_opts)]
    routine, opts = routines.pop(0)
    try:
        results = [optimize.minimize(bigauss_func, guess, args, bounds=bnds, method=routine, options=opts, tol=1e-20)]
    except ValueError:
        print('fitting error')
        import traceback
        print(traceback.format_exc())
        results = []
    while routines:
        routine, opts = routines.pop(0)
        results.append(optimize.minimize(bigauss_func, guess, args, bounds=bnds, method=routine, options=opts, tol=1e-20))
    # cdef int n = len(xdata)
    cdef float lowest = -1
    cdef np.ndarray[FLOAT_t] best
    best = results[0].x

    for i in results:
        if within_bounds(i.x, bnds):
            if lowest == -1 or i.fun < lowest:
                best = i.x
    # cdef int k = len(best.x)
    # cdef float bic = n*np.log(best.fun/n)+k+np.log(n)
    # best.bic = bic
    return best

cpdef tuple fixedMeanFit2(np.ndarray[FLOAT_t, ndim=1] xdata, np.ndarray[FLOAT_t, ndim=1] ydata,
                                       int peak_index=1, debug=False):
    cdef float rel_peak, mval
    cdef float peak_loc = xdata[peak_index]
    cdef int peak_left, peak_right
    cdef np.ndarray[FLOAT_t, ndim=1] conv_y
    cdef float peak_min, peak_max, average, variance
    cdef np.ndarray[FLOAT_t, ndim=1] bnds

    mval = ydata.max()
    conv_y = gaussian_filter1d(convolve(ydata, kaiser(10, 14), mode='same'), 3, mode='constant')
    rel_peak = conv_y[peak_index]/conv_y.max()
    ydata /= mval
    peak_left, peak_right = 0, len(xdata)#findPeak(convolve(ydata, kaiser(10, 14), mode='same'), peak_index)
    peak_min, peak_max = xdata[0], xdata[-1]
    # reset the fitting data to our bounds
    if peak_index == peak_right:
        peak_index -= peak_left-1
    else:
        peak_index -= peak_left
    if debug:
        print('left is', peak_left, 'right is', peak_right)
        print('x', xdata.tolist(), 'becomes', xdata[peak_left:peak_right].tolist())
        print('y', ydata.tolist(), 'becomes', ydata[peak_left:peak_right].tolist())
    xdata = xdata[peak_left:peak_right]
    ydata = ydata[peak_left:peak_right]
    if ydata.sum() == 0:
        return None, None
    min_spacing = min(np.diff(xdata))/2
    lb = np.fabs(peak_loc-xdata[0])
    rb = np.fabs(xdata[-1]-peak_loc)
    if lb < min_spacing:
        lb = min_spacing*5
    if rb < min_spacing:
        rb = min_spacing*5
    bnds = np.array([(rel_peak*0.75, 1.01) if rel_peak > 0 else (0.0, 1.0), (xdata[0], xdata[-1]), (min_spacing, lb), (min_spacing, rb)])
    average = np.average(xdata, weights=ydata)
    variance = np.sqrt(np.average((xdata-average)**2, weights=ydata))
    if variance == 0:
        # we have a singular peak if variance == 0, so set the variance to half of the x/y spacing
        if peak_index >= 1:
            variance = np.abs(peak_loc-xdata[peak_index-1])
        elif peak_index < len(xdata):
            variance = np.abs(xdata[peak_index+1]-peak_loc)
        else:
            # we have only 1 data point, most RT's fall into this width
            variance = 0.05
    if variance > xdata[peak_index]-peak_min or variance > peak_max-xdata[peak_index]:
        variance = xdata[peak_index]-peak_min
    if variance < min_spacing:
        variance = min_spacing
    cdef np.ndarray[FLOAT_t] guess = np.array([rel_peak, peak_loc, variance, variance])
    args = (xdata, ydata)
    base_opts = {'maxiter': 1000}
    routines = [('SLSQP', base_opts), ('TNC', base_opts), ('L-BFGS-B', base_opts)]
    routine, opts = routines.pop(0)

    if debug:
        print('guess and bounds', guess, bnds)
    try:
        results = [optimize.minimize(bigauss_func, guess, args, bounds=bnds, method=routine, options=opts, jac=bigauss_jac)]
    except ValueError:
        print('fitting error')
        import traceback
        print(traceback.format_exc())
        results = []
    while routines:# and results[-1].success == False:
        routine, opts = routines.pop(0)
        results.append(optimize.minimize(bigauss_func, guess, args, bounds=bnds, method=routine, options=opts, jac=bigauss_jac))
    # cdef int n = len(xdata)
    cdef float lowest = -1
    cdef np.ndarray[FLOAT_t] best
    if debug:
        print('fitting results', results)
    best_fit = results[0]
    for i in results:
        if within_bounds(i.x, bnds):
            if lowest == -1 or i.fun < lowest:
                best_fit = i
    best_fit.x[0]*=mval
    if debug:
        print('best fit', best_fit)
    # cdef int k = len(best.x)
    # cdef float bic = n*np.log(best.fun/n)+k+np.log(n)
    # best.bic = bic
    return best_fit.x, best_fit.fun


cdef tuple findPeak(np.ndarray[FLOAT_t, ndim=1] y, int srt):
    # check our SNR, if it's low, lessen our window
    cdef int left_offset = 1
    cdef int right_offset = 2
    cdef int lsrt = srt-left_offset if srt-left_offset > 0 else 0
    cdef int rsrt = srt+right_offset if srt+right_offset < len(y) else len(y)
    cdef float peak = y[srt]
    cdef int left = 0
    cdef np.ndarray[FLOAT_t] grad
    cdef int slope_shifts
    cdef float val
    cdef str ishift
    grad = y[lsrt:rsrt]
    try:
        shift = sum(np.sign(np.gradient(grad)))
    except ValueError:
        return 0, len(y)
    shift = 'left' if shift < 0 else 'right'
    ishift = shift
    slope_shifts = 0
    last_slope = -1
    for left in xrange(srt-1, -1, -1):
        val = y[left]
        grad = y[left-2:left+1]
        slope = None
        if len(grad) >= 2:
            slope = sum(np.sign(np.gradient(grad)))
            if slope < 0:
                slope = 'right'
            elif slope > 0:
                slope = 'left'
        if last_slope != -1:
            if last_slope != slope and slope != None:
                slope_shifts += 1
        last_slope = slope
        if ishift == 'right' and ishift == slope:
            break
        if ishift == 'left' and slope == 'right' and slope_shifts > 1:
            break
        if val == 0 or (val > peak and slope != 'right'):
            if val == 0 or shift != 'left':
                break
        elif shift == 'left' and slope != 'right':# slope != right logic: newsl
            shift = None

    cdef int right
    cdef float highest_val

    right = len(y)
    shift = ishift
    highest_val = peak
    slope_shifts = 0
    last_slope = -1
    for right in xrange(srt+1, len(y)):
        val = y[right]
        grad = y[right:right+3]
        slope = None
        if len(grad) >= 2:
            slope = sum(np.sign(np.gradient(grad)))
            if slope < 0:
                slope = 'right'
            elif slope > 0:
                slope = 'left'
        if last_slope != -1:
            if last_slope != slope and slope != None:
                slope_shifts += 1
        last_slope = slope
        if ishift == 'left' and ishift == slope:
            break
        if ishift == 'right' and slope == 'left' and slope_shifts > 1:
            break
        if val == 0 or (val > peak and slope != 'left'):
            if val > highest_val:
                highest_val = val
            if val == 0 or shift != 'right':
                if val == 0:
                    right += 1
                break
        elif shift == 'right' and slope != 'left':
            shift = None
            peak = highest_val
    return left, right

cpdef FLOAT_t get_ppm(FLOAT_t theoretical, FLOAT_t observed):
    return np.fabs(theoretical-observed)/theoretical


cpdef FLOAT_t find_nearest(np.ndarray[FLOAT_t, ndim=1] array, FLOAT_t value):
    return array[find_nearest_index(array, value)]

cpdef INT_t find_nearest_index(np.ndarray[FLOAT_t, ndim=1] array, value):
    cdef idx = np.searchsorted(array, value, side="left")
    if idx == 0:
        return 0
    elif idx == len(array):
        return -1
    elif np.fabs(value - array[idx-1]) < np.fabs(value - array[idx]):
        return idx-1
    else:
        return idx

cpdef np.ndarray[INT_t, ndim=1] find_nearest_indices(np.ndarray[FLOAT_t, ndim=1] array, np.ndarray[FLOAT_t, ndim=1] value):
    cdef np.ndarray[INT_t, ndim=1] out = np.zeros(len(value), dtype=np.int64)
    cdef np.ndarray[INT_t, ndim=1] indices = np.searchsorted(array, value, side="left")
    cdef INT_t search_index, idx
    cdef FLOAT_t search_value

    for search_index, idx in enumerate(indices):
        search_value = value[search_index]
        if idx == 0:
            out[search_index] = 0
        elif idx == len(array):
            out[search_index] = -1
        elif np.fabs(search_value - array[idx-1]) < np.fabs(search_value - array[idx]):
            out[search_index] = idx-1
        else:
            out[search_index] = idx
    return out

cpdef np.ndarray[FLOAT_t, ndim=1] gauss_hess(np.ndarray[FLOAT_t, ndim=1] params, np.ndarray[FLOAT_t, ndim=1] x, np.ndarray[FLOAT_t, ndim=1] y):
    cdef np.ndarray[FLOAT_t, ndim=1] common
    cdef np.ndarray[FLOAT_t, ndim=2] H
    H = np.zeros((params.shape[0], params.shape[0]))
    common = -1*gauss_ndim(x, params) + y
    for i in xrange(params.shape[0]):
        if i%3 == 0:
            amp = params[i]
        elif i%3 == 1:
            mu = params[i]
        elif i%3 == 2:
            sigma = params[i]
            for j in xrange(0, params.shape[0], 3):
                hess_amp, hess_mu, hess_sigma = params[j:j+3]
                # the diagonal in general refers to the hessian with partial
                # derivatives against itself (so a 3x3 or 4x4 area)
                diagonal = i == j+2
                # aa
                H[i-2, j] = sum(2 * np.exp((-(-mu+x)**2)/(sigma**2))) if diagonal else sum(2 * np.exp((-(-mu+x)**2)/(2*sigma**2)) * np.exp((-(-hess_mu+x)**2)/(2*hess_sigma**2)))
                # au
                H[i-2, j+1] = sum(-2*amp*(mu-x)/(sigma**2) * np.exp((-(-mu+x)**2)/(sigma**2)) + 2*(mu-x)/(sigma**2)*common*np.exp(-((-mu+x)**2)/(2*sigma**2))) if diagonal else sum(-2*hess_amp/(hess_sigma**2)*(hess_mu-x)*np.exp(-((-mu+x)**2)/(2*sigma**2))*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
                # as
                H[i-2, j+2] = sum(2*amp*(-mu+x)**2/(sigma**3)*np.exp(-(-mu+x)**2/(sigma**2)) - 2*((-mu+x)**2)/(sigma**3)*common*np.exp(-((-mu+x)**2)/(2*sigma**2))) if diagonal else sum(2*hess_amp*(-hess_mu+x)**2/(hess_sigma**3) * np.exp(-((-mu+x)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
                # ua
                H[i-1, j] = sum(-2*amp*(mu-x)/(sigma**2)*np.exp(-((-mu+x)**2)/(sigma**2)) + 2*(mu-x)/(sigma**2)*common*np.exp(-((-mu+x)**2)/(2*sigma**2))) if diagonal else sum(-2*amp*(mu-x)/(sigma**2)*np.exp(-((-mu+x)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))#2*amp*(mu-x)/(2*sigma**2)*np.exp(-((-mu+x)**2)/(2*sigma**2)) * 2*hess_amp*(hess_mu-x)/(2*hess_sigma**2)*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
                # uu
                H[i-1, j+1] = sum(2*amp*(amp*(mu - x)**2*np.exp(-(mu - x)**2/sigma**2)/sigma**2 + common*np.exp(-(mu - x)**2/(2*sigma**2)) - (mu - x)**2*common*np.exp(-(mu - x)**2/(2*sigma**2))/sigma**2)/sigma**2) if diagonal else sum(2*amp*(mu-x)/(2*sigma**2)*np.exp(-((-mu+x)**2)/(2*sigma**2)) * 2*hess_amp*(hess_mu-x)/(hess_sigma**2)*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
                # us
                H[i-1, j+2] = sum(-2*(amp**2)/(sigma**5)*(mu-x)*((-mu+x)**2)*np.exp(-((-mu+x)**2)/(sigma**2)) - 2*amp*2*(mu-x)/(sigma**3)*np.exp(-((-mu+x)**2)/(2*sigma**2))*common + 2*amp*(mu-x)/(sigma**5)*((-mu+x)**2)*np.exp(-((-mu+x)**2)/(2*sigma**2))*common) if diagonal else sum(-hess_amp/(hess_sigma**3)*((-hess_mu+x)**2)*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)) * 2*amp*(mu-x)/(sigma**2)*np.exp(-((-mu+x)**2)/(2*sigma**2)))
                # sa
                H[i, j] = sum(2*amp*((-mu+x)**2)/(sigma**3)*np.exp(-((-mu+x)**2)/(sigma**2)) - 2*((-mu+x)**2)/(sigma**3)*np.exp(-((-mu+x)**2)/(2*sigma**2))*common) if diagonal else sum(2*amp*((-mu+x)**2)/(sigma**3)*np.exp(-((-mu+x)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
                # su
                H[i, j+1] = sum(2*amp*(mu - x)*(-amp*(mu - x)**2*np.exp(-(mu - x)**2/sigma**2)/sigma**2 - 2*common*np.exp(-(mu - x)**2/(2*sigma**2)) + (mu - x)**2*common*np.exp(-(mu - x)**2/(2*sigma**2))/sigma**2)/sigma**3) if diagonal else sum(-2*hess_amp*(hess_mu-x)/(hess_sigma**2)*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)) * amp*(-mu+x)**2/(sigma**3)*np.exp(-((-mu+x)**2)/(2*sigma**2)))
                # ss
                H[i, j+2] = sum(2*(amp**2)*((-mu+x)**4)/(sigma**6)*np.exp(-((-mu+x)**2)/(sigma**2)) + 6*amp*((-mu+x)**2)/(sigma**4)*np.exp(-((-mu+x)**2)/(2*sigma**2))*common - 2*amp*((-mu+x)**4)/(sigma**6)*np.exp(-((-mu+x)**2)/(2*sigma**2))*common) if diagonal else sum(2*amp*((-mu+x)**2)/(sigma**3)*np.exp(-((-mu+x)**2)/(2*sigma**2)) * hess_amp*((-hess_mu+x)**2)/(hess_sigma**3)*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
    return H

# cpdef np.ndarray[FLOAT_t, ndim=1] bigauss_hess(np.ndarray[FLOAT_t, ndim=1] params, np.ndarray[FLOAT_t, ndim=1] x, np.ndarray[FLOAT_t, ndim=1] y):
#     cdef np.ndarray[FLOAT_t, ndim=1] common
#     cdef np.ndarray[FLOAT_t, ndim=2] H
#     H = np.zeros((params.shape[0], params.shape[0]))
#     common = -1*bigauss_ndim(x, params) + y
#     for i in xrange(0, params.shape[0], 4):
#         amp, mu, sigma, sigma2 = params[i:i+4]
#         for j in xrange(0, params.shape[0], 4):
#             hess_amp, hess_mu, hess_sigma, hess_sigma2 = params[j:j+4]
#             # the diagonal in general refers to the hessian with partial
#             # derivatives against itself (so a 3x3 or 4x4 area)
#             diagonal = i == j+2
#             lx = x[x<=mu]
#             left_common = common[x<=mu]
#             rx = x[x>mu]
#             right_common = common[x>mu]
#
#             hess_lx = x[x<=hess_mu]
#             hess_rx = x[x>hess_mu]
#
#             # aa left
#             H[i, j] += sum(2 * np.exp((-(-mu+lx)**2)/(sigma**2))) if diagonal else sum(2 * np.exp((-(-mu+lx)**2)/(2*sigma**2)) * np.exp((-(-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # aa right
#             H[i, j] += sum(2 * np.exp((-(-mu+rx)**2)/(sigma2**2))) if diagonal else sum(2 * np.exp((-(-mu+rx)**2)/(2*sigma2**2)) * np.exp((-(-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#             # au left
#             H[i, j+1] += sum(-2*amp*(mu-lx)/(sigma**2) * np.exp((-(-mu+lx)**2)/(sigma**2)) + 2*(mu-lx)/(sigma**2)*left_common*np.exp(-((-mu+lx)**2)/(2*sigma**2))) if diagonal else sum(-2*hess_amp/(hess_sigma**2)*(hess_mu-lx)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # au right
#             H[i, j+1] += sum(-2*amp*(mu-rx)/(sigma2**2) * np.exp((-(-mu+rx)**2)/(sigma2**2)) + 2*(mu-rx)/(sigma2**2)*right_common*np.exp(-((-mu+rx)**2)/(2*sigma2**2))) if diagonal else sum(-2*hess_amp/(hess_sigma2**2)*(hess_mu-rx)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#             # as
#             H[i, j+2] += sum(2*amp*(-mu+lx)**2/(sigma**3)*np.exp(-(-mu+lx)**2/(sigma**2)) - 2*((-mu+lx)**2)/(sigma**3)*left_common*np.exp(-((-mu+lx)**2)/(2*sigma**2))) if diagonal else sum(2*hess_amp*(-hess_mu+lx)**2/(hess_sigma**3) * np.exp(-((-mu+lx)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # as2
#             H[i, j+3] += sum(2*amp*(-mu+rx)**2/(sigma2**3)*np.exp(-(-mu+rx)**2/(sigma2**2)) - 2*((-mu+rx)**2)/(sigma2**3)*right_common*np.exp(-((-mu+rx)**2)/(2*sigma2**2))) if diagonal else sum(2*hess_amp*(-hess_mu+rx)**2/(hess_sigma2**3) * np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#
#             # ua left
#             H[i+1, j] += sum(-2*amp*(mu-lx)/(sigma**2)*np.exp(-((-mu+lx)**2)/(sigma**2)) + 2*(mu-lx)/(sigma**2)*left_common*np.exp(-((-mu+lx)**2)/(2*sigma**2))) if diagonal else sum(-2*amp*(mu-lx)/(sigma**2)*np.exp(-((-mu+lx)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # ua right
#             H[i+1, j] += sum(-2*amp*(mu-rx)/(sigma2**2)*np.exp(-((-mu+rx)**2)/(sigma2**2)) + 2*(mu-rx)/(sigma2**2)*right_common*np.exp(-((-mu+rx)**2)/(2*sigma2**2))) if diagonal else sum(-2*amp*(mu-rx)/(sigma2**2)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#             # uu left
#             H[i+1, j+1] += sum(2*amp*(amp*(mu - lx)**2*np.exp(-(mu - lx)**2/sigma**2)/sigma**2 + left_common*np.exp(-(mu - lx)**2/(2*sigma**2)) - (mu - lx)**2*left_common*np.exp(-(mu - lx)**2/(2*sigma**2))/sigma**2)/sigma**2) if diagonal else sum(2*amp*(mu-lx)/(2*sigma**2)*np.exp(-((-mu+lx)**2)/(2*sigma**2)) * 2*hess_amp*(hess_mu-lx)/(hess_sigma**2)*np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # uu right
#             H[i+1, j+1] += sum(2*amp*(amp*(mu - rx)**2*np.exp(-(mu - rx)**2/sigma2**2)/sigma2**2 + right_common*np.exp(-(mu - rx)**2/(2*sigma2**2)) - (mu - rx)**2*right_common*np.exp(-(mu - rx)**2/(2*sigma2**2))/sigma2**2)/sigma2**2) if diagonal else sum(2*amp*(mu-rx)/(2*sigma2**2)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * 2*hess_amp*(hess_mu-rx)/(hess_sigma2**2)*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#             # us
#             H[i+1, j+2] += sum(-2*(amp**2)/(sigma**5)*(mu-lx)*((-mu+lx)**2)*np.exp(-((-mu+lx)**2)/(sigma**2)) - 2*amp*2*(mu-lx)/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*left_common + 2*amp*(mu-lx)/(sigma**5)*((-mu+lx)**2)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*left_common) if diagonal else sum(-hess_amp/(hess_sigma**3)*((-hess_mu+lx)**2)*np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)) * 2*amp*(mu-lx)/(sigma**2)*np.exp(-((-mu+lx)**2)/(2*sigma**2)))
#             # us2
#             H[i+1, j+3] += sum(-2*(amp**2)/(sigma2**5)*(mu-rx)*((-mu+rx)**2)*np.exp(-((-mu+rx)**2)/(sigma2**2)) - 2*amp*2*(mu-rx)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*right_common + 2*amp*(mu-rx)/(sigma2**5)*((-mu+rx)**2)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*right_common) if diagonal else sum(-hess_amp/(hess_sigma2**3)*((-hess_mu+rx)**2)*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)) * 2*amp*(mu-rx)/(sigma2**2)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)))
#
#             # sa left
#             H[i+2, j] += sum(2*amp*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(sigma**2)) - 2*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*left_common) if diagonal else sum(2*amp*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # sa right
#             H[i+2, j] += sum(2*amp*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(sigma2**2)) - 2*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*right_common) if diagonal else sum(2*amp*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#             # su left
#             H[i+2, j+1] += sum(2*amp*(mu - lx)*(-amp*(mu - lx)**2*np.exp(-(mu - lx)**2/sigma**2)/sigma**2 - 2*left_common*np.exp(-(mu - lx)**2/(2*sigma**2)) + (mu - lx)**2*left_common*np.exp(-(mu - lx)**2/(2*sigma**2))/sigma**2)/sigma**3) if diagonal else sum(-2*hess_amp*(hess_mu-lx)/(hess_sigma**2)*np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)) * amp*(-mu+lx)**2/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2)))
#             # su right
#             H[i+2, j+1] += sum(2*amp*(mu - rx)*(-amp*(mu - rx)**2*np.exp(-(mu - rx)**2/sigma2**2)/sigma2**2 - 2*right_common*np.exp(-(mu - rx)**2/(2*sigma2**2)) + (mu - rx)**2*right_common*np.exp(-(mu - rx)**2/(2*sigma2**2))/sigma2**2)/sigma2**3) if diagonal else sum(-2*hess_amp*(hess_mu-rx)/(hess_sigma2**2)*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)) * amp*(-mu+rx)**2/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)))
#             # ss
#             H[i+2, j+2] += sum(2*(amp**2)*((-mu+lx)**4)/(sigma**6)*np.exp(-((-mu+lx)**2)/(sigma**2)) + 6*amp*((-mu+lx)**2)/(sigma**4)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*left_common - 2*amp*((-mu+lx)**4)/(sigma**6)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*left_common) if diagonal else sum(2*amp*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2)) * hess_amp*((-hess_mu+lx)**2)/(hess_sigma**3)*np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#
#             # ss2
#             H[i+2, j+3] += 0# if diagonal else sum(2*amp*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * hess_amp*((-hess_mu+rx)**2)/(hess_sigma2**3)*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#
#             # s2a left
#             H[i+3, j] += sum(2*amp*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(sigma**2)) - 2*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2))*left_common) if diagonal else sum(2*amp*((-mu+lx)**2)/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2)) * np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)))
#             # s2a right
#             H[i+3, j] += sum(2*amp*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(sigma2**2)) - 2*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*right_common) if diagonal else sum(2*amp*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#             # s2u left
#             H[i+3, j+1] += sum(2*amp*(mu - lx)*(-amp*(mu - lx)**2*np.exp(-(mu - lx)**2/sigma**2)/sigma**2 - 2*left_common*np.exp(-(mu - lx)**2/(2*sigma**2)) + (mu - lx)**2*left_common*np.exp(-(mu - lx)**2/(2*sigma**2))/sigma**2)/sigma**3) if diagonal else sum(-2*hess_amp*(hess_mu-lx)/(hess_sigma**2)*np.exp(-((-hess_mu+lx)**2)/(2*hess_sigma**2)) * amp*(-mu+lx)**2/(sigma**3)*np.exp(-((-mu+lx)**2)/(2*sigma**2)))
#             # s2u right
#             H[i+3, j+1] += sum(2*amp*(mu - rx)*(-amp*(mu - rx)**2*np.exp(-(mu - rx)**2/sigma2**2)/sigma2**2 - 2*right_common*np.exp(-(mu - rx)**2/(2*sigma2**2)) + (mu - rx)**2*right_common*np.exp(-(mu - rx)**2/(2*sigma2**2))/sigma2**2)/sigma2**3) if diagonal else sum(-2*hess_amp*(hess_mu-rx)/(hess_sigma2**2)*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)) * amp*(-mu+rx)**2/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)))
#
#             # s2s
#             H[i+3, j+2] += 0# if diagonal else sum(2*amp*((-mu+x)**2)/(sigma**3)*np.exp(-((-mu+x)**2)/(2*sigma**2)) * hess_amp*((-hess_mu+x)**2)/(hess_sigma**3)*np.exp(-((-hess_mu+x)**2)/(2*hess_sigma**2)))
#
#             # s2s2
#             H[i+3, j+3] += sum(2*(amp**2)*((-mu+rx)**4)/(sigma2**6)*np.exp(-((-mu+rx)**2)/(sigma2**2)) + 6*amp*((-mu+rx)**2)/(sigma2**4)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*right_common - 2*amp*((-mu+rx)**4)/(sigma2**6)*np.exp(-((-mu+rx)**2)/(2*sigma2**2))*right_common) if diagonal else sum(2*amp*((-mu+rx)**2)/(sigma2**3)*np.exp(-((-mu+rx)**2)/(2*sigma2**2)) * hess_amp*((-hess_mu+rx)**2)/(hess_sigma2**3)*np.exp(-((-hess_mu+rx)**2)/(2*hess_sigma2**2)))
#     return H



__author__ = "chris"
import pkg_resources  # part of setuptools
import argparse
from pythomics.proteomics import config

version = pkg_resources.require("pyquant")[0].version

description = """
This will quantify labeled peaks (such as SILAC) in ms1 spectra. It relies solely on the distance between peaks,
 which can correct for errors due to amino acid conversions.
"""

PEAK_RESOLUTION_RT_MODE = "rt"
PEAK_RESOLUTION_COMMON_MODE = "common-peak"

PEAK_FINDING_REL_MAX = "relative-max"
PEAK_FINDING_DERIVATIVE = "derivative"

PEAK_FIT_MODE_FAST = "fast"
PEAK_FIT_MODE_AVERAGE = "average"
PEAK_FIT_MODE_SLOW = "slow"

pyquant_parser = argparse.ArgumentParser(
    prog="PyQuant v{}".format(version),
    description=description,
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
pyquant_parser.add_argument("-p", help="Threads to run", type=int, default=1)
pyquant_parser.add_argument("--theo-xic", help=argparse.SUPPRESS, action="store_true")

raw_group = pyquant_parser.add_argument_group("Raw Data Parameters")
raw_group.add_argument(
    "--scan-file",
    help="The scan file(s) for the raw data. If not provided, assumed to be in the directory of the processed/tabbed/peaklist file.",
    type=argparse.FileType("r"),
    nargs="*",
)
raw_group.add_argument(
    "--scan-file-dir", help="The directory containing raw data.", type=str
)
raw_group.add_argument(
    "--precision",
    help="The precision for storing m/z values. Defaults to 6 decimal places.",
    type=int,
    default=6,
)
raw_group.add_argument(
    "--precursor-ppm",
    help="The mass accuracy for the first monoisotopic peak in ppm.",
    type=float,
    default=5,
)
raw_group.add_argument(
    "--isotope-ppm",
    help="The mass accuracy for the isotopic cluster.",
    type=float,
    default=2.5,
)
raw_group.add_argument(
    "--spread",
    help="Assume there is spread of the isotopic label.",
    action="store_true",
)

search_group = pyquant_parser.add_argument_group("Search Information")
search_group.add_argument(
    "--search-file",
    help="A search output or Proteome Discoverer msf file",
    type=argparse.FileType("rb"),
    required=False,
)
search_group.add_argument(
    "--skip",
    help="If true, skip scans with missing files in the mapping.",
    action="store_true",
)
search_group.add_argument(
    "--peptide", help="The peptide(s) to limit quantification to.", type=str, nargs="*"
)
search_group.add_argument(
    "--peptide-file",
    help="A file of peptide(s) to limit quantification to.",
    type=argparse.FileType("r"),
)
search_group.add_argument(
    "--scan", help="The scan(s) to limit quantification to.", type=str, nargs="*"
)

replicate_group = pyquant_parser.add_argument_group("Missing Value Analysis")
replicate_group.add_argument(
    "--mva", help="Analyze files in 'missing value' mode.", action="store_true"
)
replicate_group.add_argument(
    "--rt-window",
    help="The maximal deviation of a scan's retention time to be considered for analysis.",
    default=0.25,
    type=float,
)

label_group = pyquant_parser.add_argument_group("Labeling Information")
label_subgroup = label_group.add_mutually_exclusive_group()
label_subgroup.add_argument(
    "--label-scheme",
    help="The file corresponding to the labeling scheme utilized.",
    type=argparse.FileType("r"),
)
label_subgroup.add_argument(
    "--label-method",
    help="Predefined labeling schemes to use.",
    type=str,
    choices=sorted(config.LABEL_SCHEMES.keys()),
)
label_group.add_argument(
    "--reference-label",
    help="The label to use as a reference (by default all comparisons are taken).",
    type=str,
)

tsv_group = pyquant_parser.add_argument_group("Tabbed File Input")
tsv_group.add_argument(
    "--tsv",
    help="A delimited file containing scan information.",
    type=argparse.FileType("r"),
)
tsv_group.add_argument(
    "--label",
    help="The column indicating the label state of the peptide. If not found, entry assumed to be light variant.",
    default="Labeling State",
)
tsv_group.add_argument(
    "--peptide-col", help="The column indicating the peptide.", default="Peptide"
)
tsv_group.add_argument(
    "--rt", help="The column indicating the retention time.", default="Retention time"
)
tsv_group.add_argument(
    "--mz",
    help="The column indicating the MZ value of the precursor ion. This is not the MH+.",
    default="Light Precursor",
)
tsv_group.add_argument(
    "--scan-col",
    help="The column indicating the scan corresponding to the ion.",
    default="MS2 Spectrum ID",
)
tsv_group.add_argument(
    "--charge",
    help="The column indicating the charge state of the ion.",
    default="Charge",
)
tsv_group.add_argument(
    "--source",
    help="The column indicating the raw file the scan is contained in.",
    default="Raw file",
)

ion_search_group = pyquant_parser.add_argument_group("Targetted Ion Search Parameters")
ion_search_group.add_argument(
    "--msn-id",
    help="The ms level to search for the ion in. Default: 2 (ms2)",
    type=int,
    default=2,
)
ion_search_group.add_argument(
    "--msn-quant-from",
    help="The ms level to quantify values from. i.e. if we are identifying an ion in ms2, we can quantify it in ms1 (or ms2). Default: msn value-1",
    type=int,
    default=None,
)
ion_search_group.add_argument(
    "--msn-ion",
    help="M/Z values to search for in the scans. To search for multiple m/z values for a given ion, separate m/z values with a comma.",
    nargs="+",
    type=str,
)
ion_search_group.add_argument(
    "--msn-ion-rt", help="RT values each ion is expected at.", nargs="+", type=float
)
ion_search_group.add_argument(
    "--msn-peaklist",
    help="A file containing peaks to search for in the scans.",
    type=argparse.FileType("rb"),
)
ion_search_group.add_argument(
    "--msn-ppm",
    help="The error tolerance for identifying the ion(s).",
    type=float,
    default=200,
)
ion_search_group.add_argument(
    "--msn-rt-window",
    help="The range of retention times for identifying the ion(s). (ex:  7.54-9.43)",
    type=str,
    nargs="+",
)
ion_search_group.add_argument(
    "--msn-all-scans",
    help="Search for the ion across all scans (ie if you have 3 ions, you will have 3 results with one long XIC)",
    action="store_true",
)
ion_search_group.add_argument(
    "--require-all-ions",
    help="If multiple ions are set (in the style of 93.15,105.15), all ions must be found in a scan.",
    action="store_true",
)

quant_parameters = pyquant_parser.add_argument_group("Quantification Parameters")
quant_parameters.add_argument(
    "--quant-method",
    help="The process to use for quantification. Default: Integrate for ms1, sum for ms2+.",
    choices=["integrate", "sum"],
    default=None,
)
quant_parameters.add_argument(
    "--reporter-ion",
    help="Indicates that reporter ions are being used. As such, we only analyze a single scan.",
    action="store_true",
)
quant_parameters.add_argument(
    "--isotopologue-limit",
    help="How many isotopologues to quantify",
    type=int,
    default=-1,
)
quant_parameters.add_argument(
    "--overlapping-labels",
    help="This declares the mz values of labels will overlap. It is useful for data such as neucode, but not needed for only SILAC labeling.",
    action="store_true",
)
quant_parameters.add_argument(
    "--labels-needed",
    help="How many labels need to be detected to quantify a scan (ie if you have a 2 state experiment and set this to 2, it will only quantify scans where both occur.",
    default=1,
    type=int,
)
quant_parameters.add_argument(
    "--merge-labels", help="Merge labels together to a single XIC.", action="store_true"
)
quant_parameters.add_argument(
    "--min-scans",
    help="How many quantification scans are needed to quantify a scan.",
    default=1,
    type=int,
)
quant_parameters.add_argument(
    "--min-resolution",
    help="The minimal resolving power of a scan to consider for quantification. Useful for skipping low-res scans",
    default=0,
    type=float,
)
quant_parameters.add_argument(
    "--no-mass-accuracy-correction",
    help="Disables the mass accuracy correction.",
    action="store_true",
)
quant_parameters.add_argument(
    "--no-contaminant-detection",
    help="Disables routine to check if an ion is a contaminant of a nearby peptide (checks if its a likely isotopologue).",
    action="store_true",
)

peak_parameters = pyquant_parser.add_argument_group("Peak Fitting Parameters")
peak_parameters.add_argument(
    "--peak-find-method",
    help="The method to use to identify peaks within data. For LC-MS, relative-max is usually best. For smooth data, derivative is better.",
    type=str,
    choices=(PEAK_FINDING_REL_MAX, PEAK_FINDING_DERIVATIVE),
    default=PEAK_FINDING_REL_MAX,
)
peak_parameters.add_argument(
    "--peak-find-mode",
    help="This picks some predefined parameters for various use cases. Fast is good for robust data with few peaks, slow is good for complex data with overlapping peaks of very different size.",
    type=str,
    choices=(PEAK_FIT_MODE_SLOW, PEAK_FIT_MODE_AVERAGE, PEAK_FIT_MODE_FAST),
    default=PEAK_FIT_MODE_AVERAGE,
)
peak_parameters.add_argument(
    "--gap-interpolation",
    help="This interpolates missing data in scans. The parameter should be a number that is the maximal gap size to fill (ie 2 means a gap of 2 scans). Can be useful for low intensity LC-MS data.",
    type=int,
    default=0,
)
peak_parameters.add_argument(
    "--fit-baseline",
    help="Fit a separate line for the baseline of each peak.",
    action="store_true",
)
peak_parameters.add_argument(
    "--peak-cutoff",
    help="The threshold from the initial retention time a peak can fall by before being discarded",
    type=float,
    default=0.05,
)
peak_parameters.add_argument(
    "--max-peaks",
    help="The maximal number of peaks to detect per scan. A lower value can help with very noisy data.",
    type=int,
    default=-1,
)
peak_parameters.add_argument(
    "--peaks-n",
    help="The number of peaks to report per scan. Useful for ions with multiple elution times.",
    type=int,
    default=1,
)
peak_parameters.add_argument(
    "--no-rt-guide",
    help="Do not use the retention time to bias for peaks containing the MS trigger time.",
    action="store_true",
)
peak_parameters.add_argument(
    "--snr-filter", help="Filter peaks below a given SNR.", type=float, default=0
)
peak_parameters.add_argument(
    "--zscore-filter",
    help="Peaks below a given z-score are excluded.",
    type=float,
    default=0,
)
peak_parameters.add_argument(
    "--filter-width",
    help="The window size for snr/zscore filtering. Default: entire scan",
    type=float,
    default=0,
)
peak_parameters.add_argument(
    "--r2-cutoff",
    help="The minimal R^2 for a peak to be kept. Should be a value between 0 and 1",
    type=float,
    default=None,
)
peak_parameters.add_argument(
    "--intensity-filter",
    help="Filter peaks whose peak are below a given intensity.",
    type=float,
    default=0,
)
peak_parameters.add_argument(
    "--percentile-filter",
    help="Filter peaks whose peak are below a given percentile of the data.",
    type=float,
    default=0,
)
peak_parameters.add_argument(
    "--min-peak-separation",
    help="Peaks separated by less than this distance will be combined. For very crisp data, set this to a lower number. (minimal value is 1)",
    type=int,
    default=5,
)
peak_parameters.add_argument(
    "--disable-peak-filtering",
    help="This will disable smoothing of data prior to peak finding. If you have very good LC, this may be used to identify small peaks.",
    action="store_true",
)
peak_parameters.add_argument(
    "--merge-isotopes",
    help="Merge Isotopologues together prior to fitting.",
    action="store_true",
)
peak_parameters.add_argument(
    "--peak-resolution-mode",
    help="The method to use to resolve peaks across multiple XICs",
    choices=(PEAK_RESOLUTION_RT_MODE, PEAK_RESOLUTION_COMMON_MODE),
    type=str,
    default="common-peak",
)

# Deprecated parameters
peak_parameters.add_argument(
    "--remove-baseline",
    help=argparse.SUPPRESS,
    action="store_true",
    dest="fit_baseline",
)


xic_parameters = pyquant_parser.add_argument_group("XIC Options")
xic_parameters.add_argument(
    "--xic-snr",
    help="When the SNR of the XIC falls below this, stop searching for more data. Useful for escaping from noisy shoulders and contaminants.",
    type=float,
    default=1.0,
)
xic_parameters.add_argument(
    "--xic-missing-ion-count",
    help="This specifies how many consequtive scans an ion can be missing for until it is no longer considered.",
    type=int,
    default=1,
)
xic_parameters.add_argument(
    "--xic-window-size",
    help="When the number of scans in a given direction from the initial datapoint of an XIC passes this, stop. Default is -1 (disabled). Useful for removing contaminants",
    type=int,
    default=-1,
)
xic_parameters.add_argument(
    "--xic-smooth",
    help="Prior to fitting, smooth data with a Gaussian filter.",
    action="store_true",
)
xic_parameters.add_argument(
    "--export-msn",
    help="This will export spectra of a given MSN that were used to provide the quantification.",
    action="store_false",
)


mrm_parameters = pyquant_parser.add_argument_group("SRM/MRM Parameters")
#'A file indicating light and heavy peptide pairs, and optionally the known elution time.'
mrm_parameters.add_argument(
    "--mrm-map", help=argparse.SUPPRESS, type=argparse.FileType("r")
)

output_group = pyquant_parser.add_argument_group("Output Options")
output_group.add_argument(
    "--debug", help="This will output debug information.", action="store_true"
)
output_group.add_argument(
    "--html", help="Output a HTML table summary.", action="store_true"
)
output_group.add_argument(
    "--resume",
    help="Will resume from the last run. Only works if not directing output to stdout.",
    action="store_true",
)
output_group.add_argument(
    "--sample",
    help="How much of the data to sample. Enter as a decimal (ie 1.0 for everything, 0.1 for 10%%)",
    type=float,
    default=1.0,
)
output_group.add_argument(
    "--disable-stats",
    help="Disable confidence statistics on data.",
    action="store_true",
)
output_group.add_argument(
    "--no-ratios", help="Disable reporting of ratios in output.", action="store_true"
)
output_group.add_argument(
    "-o", "--out", nargs="?", help="The prefix for the file output", type=str
)

PER_PEAK = "per-peak"
PER_FILE = "per-file"
PER_ID = "per-id"

spectra_output = pyquant_parser.add_argument_group("Spectra Output Options")
spectra_output.add_argument(
    "--export-mzml",
    help="Create an mzml file of spectra contained within each peak.",
    action="store_true",
)
spectra_output.add_argument(
    "--export-mode",
    help="How to export the scans. per-peak: A mzML per peak identified. per-id: A mzML per ion identified (each row of the output gets an mzML). per-file: All scans matched per raw file.",
    type=str,
    default="per-peak",
    choices={PER_PEAK, PER_ID, PER_FILE},
)

convenience_group = pyquant_parser.add_argument_group("Convenience Parameters")
convenience_group.add_argument(
    "--neucode",
    help="This will select parameters specific for neucode. Note: You still must define a labeling scheme.",
    action="store_true",
)
convenience_group.add_argument(
    "--isobaric-tags",
    help="This will select parameters specific for isobaric tag based labeling (TMT/iTRAQ).",
    action="store_true",
)
convenience_group.add_argument(
    "--ms3",
    help="This will select parameters specific for ms3 based quantification.",
    action="store_true",
)
convenience_group.add_argument(
    "--maxquant",
    help="This will select parameters specific for a MaxQuant evidence file.",
    action="store_true",
)
convenience_group.add_argument(
    "--gcms",
    help="This will select parameters specific for ion identification and quantification in GCMS experiments.",
    action="store_true",
)
#'This will select parameters specific for Selective/Multiple Reaction Monitoring (SRM/MRM).'
convenience_group.add_argument("--mrm", help=argparse.SUPPRESS, action="store_true")



import os
import subprocess
from unittest import TestCase

import numpy as np
import pandas as pd

from .mixins import FileMixins
from . import config


class TestConvenience(FileMixins, TestCase):
    def setUp(self):
        super(TestConvenience, self).setUp()
        self.output = os.path.join(self.out_dir, "mq")
        self.mq_file = os.path.join(self.data_dir, "ecoli_mq_124_ms2.txt")

    def test_mq_convenience(self):
        com = [
            self.executable,
            "--scan-file",
            self.ecoli_mzml,
            "-p",
            str(config.CORES),
            "-o",
            self.output,
            "--maxquant",
            "--tsv",
            self.mq_file,
            "--html",
            "--label-method",
            "K4K8R6R10",
            "--sample",
            "0.1",
            "--gap-interpolation",
            "2",
        ]
        subprocess.call(com)
        pyquant = pd.read_table(self.output)
        label = "Heavy"
        pq_sel = "{}/Light".format(label)
        pyquant[pq_sel] = np.log2(pyquant[pq_sel] + 0.000001)
        r10_med = pyquant[pq_sel].median()
        self.assertNotAlmostEqual(r10_med, -2.0, places=2)



__author__ = "chris"
import unittest
from random import randint, random

import numpy as np
from sympy import symbols, diff, exp, Piecewise

from .mixins import GaussianMixin
from pyquant import peaks
from pyquant import cpeaks


def get_gauss_value(x, amp, mu, std):
    return amp * np.exp(-((x - mu) ** 2) / (2 * std ** 2))


class MathTests(GaussianMixin, unittest.TestCase):
    def setUp(self):
        super(MathTests, self).setUp()
        self.std_2, self.std2_2 = 0.5, 0.75
        self.one_bigauss_params = np.array(
            [self.amp, self.mu, self.std, self.std_2], dtype=np.float
        )
        self.two_bigauss_params = np.array(
            [
                self.amp,
                self.mu,
                self.std,
                self.std_2,
                self.amp,
                self.mu2,
                self.std,
                self.std2_2,
            ],
            dtype=np.float,
        )
        self.one_bigauss = peaks.bigauss_ndim(self.x, self.one_bigauss_params)
        self.two_bigauss = peaks.bigauss_ndim(self.x, self.two_bigauss_params)

    def test_jacobians(self):
        one_gauss_jac = peaks.gauss_jac(
            self.one_gauss_params, self.x, self.one_gauss, False
        )
        self.assertEqual(
            one_gauss_jac.tolist(), np.zeros_like(self.one_gauss_params).tolist()
        )

        two_gauss_jac = peaks.gauss_jac(
            self.two_gauss_params, self.x, self.two_gauss, False
        )
        self.assertEqual(
            two_gauss_jac.tolist(), np.zeros_like(self.two_gauss_params).tolist()
        )
        one_bigauss_jac = peaks.bigauss_jac(
            self.one_bigauss_params, self.x, self.one_bigauss, False
        )
        self.assertEqual(
            one_bigauss_jac.tolist(), np.zeros_like(self.one_bigauss_params).tolist()
        )

        two_bigauss_jac = peaks.bigauss_jac(
            self.two_bigauss_params, self.x, self.two_bigauss, False
        )
        self.assertEqual(
            two_bigauss_jac.tolist(), np.zeros_like(self.two_bigauss_params).tolist()
        )
        y, x, a, u, s1, s1_2, a2, u2, s2, s2_2, a3, u3, s3, s3_2 = symbols(
            "y x a u s1 s1_2 a2 u2 s2 s2_2 a3 u3 s3 s3_2"
        )
        three_gauss = (
            y
            - (
                a * exp(-((u - x) ** 2) / (2 * s1 ** 2))
                + a2 * exp(-((u2 - x) ** 2) / (2 * s2 ** 2))
                + a3 * exp(-((u3 - x) ** 2) / (2 * s3 ** 2))
            )
        ) ** 2
        three_gauss2 = (
            y
            - (
                a * exp(-((u - x) ** 2) / (2 * s1_2 ** 2))
                + a2 * exp(-((u2 - x) ** 2) / (2 * s2_2 ** 2))
                + a3 * exp(-((u3 - x) ** 2) / (2 * s3_2 ** 2))
            )
        ) ** 2
        bigauss = Piecewise((three_gauss, x < u))
        deriv_store = {}
        for i in range(2):
            subs = [
                ("a", random()),
                ("u", randint(1, 10)),
                ("s1", random()),
                ("a2", random()),
                ("u2", randint(12, 20)),
                ("s2", random()),
                ("a3", random()),
                ("u3", randint(22, 30)),
                ("s3", random()),
            ]
            params = np.array([i[1] for i in subs], dtype=float)
            noisy_params = params + 2 * np.random.rand(params.shape[0])
            gauss_x = np.linspace(-10, 40, 100)
            gauss_y = peaks.gauss_ndim(gauss_x, noisy_params)
            jacobian = peaks.gauss_jac(params, gauss_x, gauss_y, False)
            for var_index, var in enumerate([a, u, s1, a2, u2, s2, a3, u3, s3]):
                deriv = deriv_store.setdefault(var, diff(three_gauss, var))
                pq_jac = jacobian[var_index]
                sympy_jacobian = sum(
                    [
                        deriv.subs(dict(subs, **{"x": xi, "y": yi}))
                        for xi, yi in zip(gauss_x, gauss_y)
                    ]
                )
                np.testing.assert_allclose(
                    pq_jac,
                    np.array(sympy_jacobian, dtype=float),
                    err_msg="d{} - pq: {}, sympy: {}".format(
                        var, pq_jac, sympy_jacobian
                    ),
                    atol=1e-4,
                )

    def test_hessians(self):
        y, x, a, u, s1, a2, u2, s2, a3, u3, s3 = symbols("y x a u s1 a2 u2 s2 a3 u3 s3")
        three_gauss = (
            y
            - (
                a * exp(-((u - x) ** 2) / (2 * s1 ** 2))
                + a2 * exp(-((u2 - x) ** 2) / (2 * s2 ** 2))
                + a3 * exp(-((u3 - x) ** 2) / (2 * s3 ** 2))
            )
        ) ** 2
        hess_store = {}
        for _ in range(2):
            subs = [
                ("a", random() + 0.5),
                ("u", random() * 10),
                ("s1", random() * 3 + 2),
                ("a2", random() + 0.5),
                ("u2", random() * 10 + 10),
                ("s2", random() * 3 + 2),
                ("a3", random() + 0.5),
                ("u3", random() * 10 + 20),
                ("s3", random() * 3 + 2),
            ]
            params = np.array([i[1] for i in subs], dtype=float)
            noisy_params = params + 2 * np.random.rand(params.shape[0])
            gauss_x = np.linspace(-10, 40, 100)
            gauss_y = cpeaks.gauss_ndim(gauss_x, noisy_params)
            hessian = cpeaks.gauss_hess(params, gauss_x, gauss_y)
            for var_index, var in enumerate([a, u, s1, a2, u2, s2, a3, u3, s3]):
                for var_index2, var2 in enumerate([a, u, s1, a2, u2, s2, a3, u3, s3]):
                    deriv = hess_store.setdefault(
                        (var, var2), diff(three_gauss, var, var2)
                    )
                    sympy_hessian = sum(
                        [
                            deriv.subs(dict(subs, **{"x": xi, "y": yi}))
                            for xi, yi in zip(gauss_x, gauss_y)
                        ]
                    )
                    pq_hess = hessian[var_index, var_index2]
                    np.testing.assert_allclose(
                        pq_hess,
                        np.array(sympy_hessian, dtype=float),
                        err_msg="d{}d{} - pq: {}, sympy: {}".format(
                            var, var2, pq_hess, sympy_hessian
                        ),
                        atol=1e-4,
                    )


if __name__ == "__main__":
    unittest.main()



import os
import sys
from collections import defaultdict, OrderedDict
from copy import deepcopy
from operator import itemgetter, attrgetter

import numpy as np
from pythomics.proteomics.config import CARBON_NEUTRON
from scipy import optimize
from scipy.signal import convolve, kaiser

if os.environ.get("PYQUANT_DEV", False) == "True":
    try:
        import pyximport

        pyximport.install(
            setup_args={"include_dirs": np.get_include()}, reload_support=True
        )
    except Exception as e:
        import traceback

        traceback.print_exc()
        pass

from .cpeaks import (
    bigauss_func,
    gauss_func,
    bigauss_ndim,
    gauss_ndim,
    bigauss_jac,
    gauss_jac,
    find_nearest,
    find_nearest_index,
    get_ppm,
)  # noqa: E402
from . import (
    PEAK_FINDING_REL_MAX,
    PEAK_FIT_MODE_AVERAGE,
    PEAK_FIT_MODE_FAST,
    PEAK_FIT_MODE_SLOW,
)  # noqa: E402
from .logger import logger  # noqa: E402
from .utils import (
    divide_peaks,
    find_possible_peaks,
    estimate_peak_parameters,
    interpolate_data,
    savgol_smooth,
    subtract_baseline,
)  # noqa: E402


_epsilon = np.sqrt(np.finfo(float).eps)


def findEnvelope(
    xdata,
    ydata,
    measured_mz=None,
    theo_mz=None,
    max_mz=None,
    precursor_ppm=5,
    isotope_ppm=2.5,
    isotope_ppms=None,
    charge=2,
    debug=False,
    isotope_offset=0,
    isotopologue_limit=-1,
    theo_dist=None,
    label=None,
    skip_isotopes=None,
    last_precursor=None,
    quant_method="integrate",
    reporter_mode=False,
    fragment_scan=False,
    centroid=False,
    contaminant_search=True,
):
    # returns the envelope of isotopic peaks as well as micro envelopes  of each individual cluster
    spacing = CARBON_NEUTRON / float(charge)
    start_mz = (
        measured_mz
        if isotope_offset == 0
        else measured_mz + isotope_offset * CARBON_NEUTRON / float(charge)
    )
    initial_mz = start_mz
    if max_mz is not None:
        max_mz = (
            max_mz - spacing * 0.9
            if isotope_offset == 0
            else max_mz + isotope_offset * CARBON_NEUTRON * 0.9 / float(charge)
        )
    if isotope_ppms is None:
        isotope_ppms = {}
    tolerance = isotope_ppms.get(0, precursor_ppm) / 1000000.0
    env_dict, micro_dict, ppm_dict = OrderedDict(), OrderedDict(), OrderedDict()
    empty_dict = {"envelope": env_dict, "micro_envelopes": micro_dict, "ppms": ppm_dict}

    non_empty = xdata[ydata > 0]
    if len(non_empty) == 0:
        if debug:
            print("data is empty")
        return empty_dict
    first_mz = find_nearest(non_empty, start_mz)

    isotope_index = 0
    use_theo = False
    # This is purposefully verbose to be more explicit
    if reporter_mode is False and fragment_scan is False:
        while get_ppm(start_mz, first_mz) > tolerance:
            # let's try using our theoretical mass
            first_mz = find_nearest(non_empty, theo_mz)
            if get_ppm(theo_mz, first_mz) > tolerance:
                # let's check our last boundary. This allows for drift in m/z values
                # as scans progress instead of enforcing the m/z at the first
                # observed instance of a given m/z
                if last_precursor is not None:
                    first_mz = find_nearest(non_empty, last_precursor)
                    if get_ppm(last_precursor, first_mz) > tolerance:
                        # repeat all of that for the next isotopic index
                        start_mz += spacing
                        initial_mz += spacing
                        theo_mz += spacing
                        last_precursor += spacing
                        isotope_index += 1
                    else:
                        start_mz = last_precursor
                        break
                else:
                    start_mz += spacing
                    theo_mz += spacing
                    initial_mz += spacing
                    isotope_index += 1
            else:
                use_theo = True
                break
            tolerance = isotope_ppms.get(isotope_index, isotope_ppm) / 1000000.0
            if isotope_index == 2 or (max_mz is not None and first_mz >= max_mz):
                if debug:
                    print("unable to find start ion")
                return empty_dict

    precursor_tolerance = tolerance

    isotope_index += isotope_offset
    start_index = find_nearest_index(xdata, first_mz)
    start_info = findMicro(
        xdata,
        ydata,
        start_index,
        ppm=tolerance,
        start_mz=start_mz,
        calc_start_mz=theo_mz,
        quant_method=quant_method,
        reporter_mode=reporter_mode,
        fragment_scan=fragment_scan,
        centroid=centroid,
    )
    start_error = start_info["error"]

    if "params" in start_info:
        if fragment_scan is False and start_info["error"] > tolerance:
            start = (
                last_precursor
                if last_precursor is not None
                else theo_mz
                if use_theo
                else start_mz
            )
        else:
            start = start_info["params"][1]
    else:
        if debug:
            print("empty start info", start_info)
        return empty_dict

    valid_locations2 = OrderedDict()
    valid_locations2[isotope_index] = [(0, start, find_nearest_index(non_empty, start))]
    contaminant_int = 0.0

    if not reporter_mode and (
        isotopologue_limit == -1 or len(valid_locations2) < isotopologue_limit
    ):
        isotope_index += 1
        pos = find_nearest_index(non_empty, start) + 1
        offset = isotope_index * spacing
        df_len = non_empty.shape[0]
        last_displacement = None
        valid_locations = []

        # check for contaminant at doubly and triply charged positions to see if we're in another ion's peak
        if contaminant_search:
            for i in range(2, 4):
                closest_contaminant = find_nearest(
                    non_empty, start - CARBON_NEUTRON / float(i)
                )
                closest_contaminant_index = find_nearest_index(
                    xdata, closest_contaminant
                )
                contaminant_bounds = findMicro(
                    xdata,
                    ydata,
                    closest_contaminant_index,
                    ppm=precursor_tolerance,
                    calc_start_mz=start,
                    start_mz=start,
                    isotope=-1,
                    spacing=CARBON_NEUTRON / float(i),
                    quant_method=quant_method,
                    centroid=centroid,
                )
                if contaminant_bounds.get("int", 0) > contaminant_int:
                    contaminant_int = contaminant_bounds.get("int", 0.0)

        # set the tolerance for isotopes
        tolerance = isotope_ppms.get(isotope_index, isotope_ppm) / 1000000.0

        while pos < df_len:
            # search for the ppm error until it rises again, we select the minima and if this minima is
            # outside our ppm error, we stop the expansion of our isotopic cluster
            current_loc = non_empty[pos]
            if max_mz is not None and current_loc >= max_mz:
                if not valid_locations:
                    break
                displacement = (
                    last_displacement + tolerance
                    if last_displacement is not None
                    else tolerance * 2
                )
            else:
                displacement = get_ppm(start + offset, current_loc)
            # because the peak location may be between two readings, we use a very tolerance search here and enforce the ppm at the peak fitting stage.
            if displacement < tolerance * 5:
                valid_locations.append((displacement, current_loc, pos))
            if last_displacement is not None:
                if valid_locations and displacement > last_displacement:
                    # pick the peak closest to our error tolerance
                    valid_locations2[isotope_index] = valid_locations
                    isotope_index += 1
                    tolerance = isotope_ppms.get(isotope_index, isotope_ppm) / 1000000.0
                    offset = spacing * isotope_index
                    displacement = get_ppm(start + offset, current_loc)
                    valid_locations = []
                    if isotopologue_limit != -1 and (
                        len(valid_locations2) >= isotopologue_limit
                    ):
                        break
                elif displacement > last_displacement and not valid_locations:
                    break
            last_displacement = displacement
            pos += 1

    # combine any overlapping micro envelopes
    valid_keys = sorted(
        set(valid_locations2.keys()).intersection(
            theo_dist.keys() if theo_dist is not None else valid_locations2.keys()
        )
    )
    best_locations = [
        sorted(valid_locations2[i], key=itemgetter(0))[0] for i in valid_keys
    ]

    for index, isotope_index in enumerate(valid_keys):
        if skip_isotopes is not None and isotope_index in skip_isotopes:
            continue
        _, _, empty_index = best_locations[index]
        micro_index = find_nearest_index(xdata, non_empty[empty_index])
        if ydata[micro_index] == 0:
            micro_index += 1
        if ydata[micro_index] == 0:
            micro_index -= 2
        # if micro_index == 0:
        #     pass
        isotope_tolerance = isotope_ppms.get(isotope_index, isotope_ppm) / 1000000.0
        micro_bounds = findMicro(
            xdata,
            ydata,
            micro_index,
            ppm=precursor_tolerance if isotope_index == 0 else isotope_tolerance,
            calc_start_mz=start,
            start_mz=start_mz,
            isotope=isotope_index,
            spacing=spacing,
            quant_method=quant_method,
            centroid=centroid,
        )
        if isotope_index == 0:
            micro_bounds["error"] = start_error

        micro_dict[isotope_index] = micro_bounds
        env_dict[isotope_index] = micro_index
        ppm_dict[isotope_index] = micro_bounds.get("error")

    # in all cases, the envelope is going to be either monotonically decreasing, or a parabola (-x^2)
    isotope_pattern = [
        (isotope_index, isotope_dict["int"])
        for isotope_index, isotope_dict in micro_dict.items()
    ]
    # Empirically, it's been found that enforcing the theoretical distribution on a per ms1 scan basis leads to
    # significant increases in variance for the XIC, so don't do it here
    if contaminant_int > 1:
        for i, (isotope_index, isotope_intensity) in enumerate(isotope_pattern):
            if contaminant_int > isotope_intensity:
                if debug:
                    print("contaminant loss", label)
                env_dict.pop(isotope_index)
                micro_dict.pop(isotope_index)
                ppm_dict.pop(isotope_index)
    # rebuild the pattern after contaminants are removed
    isotope_pattern = [
        (isotope_index, isotope_dict["int"])
        for isotope_index, isotope_dict in micro_dict.items()
    ]
    if theo_dist is None:
        # are we monotonically decreasing?
        remove = False
        if len(isotope_pattern) > 2:
            # check if the 2nd isotope is smaller than the first. This is a classical case looking like:
            #
            #  |
            #  |  |
            #  |  |  |
            #  |  |  |  |

            if isotope_pattern[1][1] < isotope_pattern[0][1]:
                # we are, check this trend holds and remove isotopes it fails for
                for i, j in zip(isotope_pattern, isotope_pattern[1:]):
                    if j[1] * 0.9 > i[1]:
                        # the pattern broke, remove isotopes beyond this point
                        remove = True
                    if remove:
                        if debug:
                            print("pattern2.1 loss", label, j[0], isotope_pattern)
                        env_dict.pop(j[0])
                        micro_dict.pop(j[0])
                        ppm_dict.pop(j[0])

            # check if the 2nd isotope is larger than the first. This is a case looking like:
            #
            #
            #     |  |
            #     |  |
            #  |  |  |  |

            elif isotope_pattern[1][1] > isotope_pattern[0][1]:
                shift = False
                for i, j in zip(isotope_pattern, isotope_pattern[1:]):
                    if shift and j[1] * 0.9 > i[1]:
                        remove = True
                    elif shift is False and j[1] < i[1] * 0.9:
                        if shift:
                            remove = True
                        else:
                            shift = True
                    if remove:
                        if debug:
                            print("pattern2.2 loss", label, j[0], isotope_pattern)
                        env_dict.pop(j[0])
                        micro_dict.pop(j[0])
                        ppm_dict.pop(j[0])

    return {
        "envelope": env_dict,
        "micro_envelopes": micro_dict,
        "ppms": ppm_dict,
    }


def findAllPeaks(
    xdata,
    ydata_original,
    min_dist=0,
    method=None,
    local_filter_size=0,
    filter=False,
    peak_boost=False,
    bigauss_fit=False,
    rt_peak=None,
    mrm=False,
    max_peaks=4,
    debug=False,
    peak_width_start=3,
    snr=0,
    zscore=0,
    amplitude_filter=0,
    peak_width_end=4,
    fit_baseline=False,
    rescale=True,
    fit_negative=False,
    percentile_filter=0,
    micro=False,
    method_opts=None,
    smooth=False,
    r2_cutoff=None,
    peak_find_method=PEAK_FINDING_REL_MAX,
    min_slope=None,
    min_peak_side_width=3,
    gap_interpolation=0,
    min_peak_width=None,
    min_peak_increase=None,
    chunk_factor=0.1,
    fit_mode=PEAK_FIT_MODE_AVERAGE,
    baseline_subtraction=False,
    **kwargs
):

    # Deprecation things
    if "baseline_correction" in kwargs:
        fit_baseline = kwargs["baseline_correction"]

    if micro:
        fit_baseline = False

    if not micro and gap_interpolation:
        ydata_original = interpolate_data(
            xdata, ydata_original, gap_limit=gap_interpolation
        )

    rel_peak_constraint = 0.0 if fit_baseline else 0.5
    original_max = (
        np.abs(ydata_original).max() if fit_negative else ydata_original.max()
    )
    amplitude_filter /= original_max
    ydata = ydata_original / original_max
    ydata_peaks = np.copy(ydata)

    if baseline_subtraction:
        ydata_peaks = subtract_baseline(ydata_peaks)

    if smooth and len(ydata) > 5:
        ydata_peaks = savgol_smooth(ydata_peaks)

    if filter or peak_boost:
        if len(ydata) >= 5:
            ydata_peaks = convolve(ydata_peaks, kaiser(10, 12), mode="same")

    ydata_peaks[np.isnan(ydata_peaks)] = 0

    ydata_peaks /= np.abs(ydata_peaks).max() if fit_negative else ydata_peaks.max()

    final_peaks = find_possible_peaks(
        xdata,
        ydata,
        ydata_peaks,
        peak_find_method=peak_find_method,
        peak_width_start=peak_width_start,
        peak_width_end=peak_width_end,
        snr=snr,
        zscore=zscore,
        rt_peak=rt_peak,
        amplitude_filter=amplitude_filter,
        fit_negative=fit_negative,
        percentile_filter=percentile_filter,
        local_filter_size=local_filter_size,
        micro=micro,
        min_slope=min_slope,
        min_dist=min_dist,
        min_peak_side_width=min_peak_side_width,
        min_peak_width=min_peak_width,
        smooth=smooth,
        min_peak_increase=min_peak_increase,
    )

    # Next, for fitting multiple peaks, we want to divide up the space so we are not fitting peaks that
    # have no chance of actually impacting one another.
    CHUNK_MAP = {
        PEAK_FIT_MODE_SLOW: 0.1,
        PEAK_FIT_MODE_AVERAGE: 0.5,
        PEAK_FIT_MODE_FAST: 1.0,
    }
    chunks = divide_peaks(
        np.abs(ydata_peaks),
        min_sep=5 if 5 > peak_width_end else peak_width_end,
        chunk_factor=CHUNK_MAP[fit_mode],
    )
    if not chunks.any() or chunks[-1] != len(ydata_peaks):
        chunks = np.hstack((chunks, len(ydata_peaks)))

    logger.debug("found: {}\n".format(final_peaks))

    step_size = 4 if bigauss_fit else 3
    if fit_baseline:
        step_size += 2
    min_spacing = min(np.diff(xdata)) / 2
    peak_range = xdata[-1] - xdata[0]
    # initial bound setup
    initial_bounds = [(-1.01, 1.01), (xdata[0], xdata[-1]), (min_spacing, peak_range)]
    if bigauss_fit:
        initial_bounds.extend([(min_spacing, peak_range)])
    if fit_baseline:
        initial_bounds.extend([(None, None), (None, None)])
        # print(final_peaks)
    if debug:
        sys.stderr.write("final peaks: {}\n".format(final_peaks))

    fitted_segments = defaultdict(list)
    for peak_width, peak_info in final_peaks.items():
        row_peaks = peak_info["peaks"]
        if not row_peaks.any():
            continue

        minima_array = np.array(peak_info["minima"], dtype=np.long)
        # Now that we have estimated the parameters for fitting all the data, we divide it up into
        # chunks and fit each segment. The choice to fit all parameters first is to prevent cases
        # where a chunk is dividing two overlapping points and the variance estimate may be too low.
        for chunk_index, right_break_point in enumerate(chunks):
            left_break_point = chunks[chunk_index - 1] if chunk_index != 0 else 0
            segment_x = xdata[left_break_point:right_break_point]
            segment_y = deepcopy(ydata[left_break_point:right_break_point])

            if not segment_y.any():
                continue

            if not micro:
                segment_max = np.abs(segment_y).max()
                segment_y /= segment_max

            segment_row_peaks = []
            segment_minima_array = []
            for i in row_peaks:
                if i >= left_break_point:
                    if i < right_break_point:
                        segment_row_peaks.append(i - left_break_point)
                    else:
                        break

            for i in minima_array:
                if i >= left_break_point:
                    if i < right_break_point:
                        segment_minima_array.append(i - left_break_point)
                    else:
                        break

            if not segment_row_peaks:
                continue

            # Get peak parameter estimates and boundaries for this segment
            segment_guess, segment_bounds = estimate_peak_parameters(
                segment_x,
                segment_y,
                np.array(segment_row_peaks),
                np.array(segment_minima_array),
                fit_negative=fit_negative,
                rel_peak_constraint=rel_peak_constraint,
                micro=micro,
                bigauss_fit=bigauss_fit,
                fit_baseline=fit_baseline,
            )

            if not segment_guess:
                continue

            args = (segment_x, segment_y, fit_baseline)
            opts = method_opts or {"maxiter": 1000}

            # Because the amplitude of peaks can vary wildly, we have to make sure our tolerance matters for the
            # smallest peaks. i.e. if we are fitting two peaks, one with an amplitude of 20M and another with 10000,
            # changes in the smaller peak will be below our tolerance and the minimization routine can ignore them

            if "ftol" not in opts and not micro:
                min_tol = 1e-10
                for i, j in zip(segment_bounds, segment_guess):
                    abs_i = np.abs(i[0]) if i[0] else None
                    abs_j = np.abs(j) if j else None
                    if abs_i and abs_i < min_tol:
                        min_tol = abs_i / 5.0
                    if abs_j and abs_j < min_tol:
                        min_tol = abs_j / 5.0
                opts["ftol"] = min_tol

            fit_func = bigauss_func if bigauss_fit else gauss_func

            routines = ["SLSQP", "TNC", "L-BFGS-B"]
            if method:
                routines = [method]

            routine = routines.pop(0)
            if len(segment_bounds) == 0:
                segment_bounds = deepcopy(initial_bounds)

            # Check that the bounds for the mean are within the segment so the optimizer doesn't try and cheat
            # by going to solutions outside of the data
            if not micro:
                for i in range(1, len(segment_bounds), step_size):
                    if segment_bounds[i][0] < segment_x[0]:
                        segment_bounds[i] = (segment_x[0], segment_bounds[i][1])
                    if segment_bounds[i][1] > segment_x[-1]:
                        segment_bounds[i] = (segment_bounds[i][0], segment_x[-1])

            if fit_baseline:
                jacobian = None
            else:
                jacobian = bigauss_jac if bigauss_fit else gauss_jac

            if debug:
                print(
                    "left and right segments",
                    xdata[left_break_point],
                    xdata[right_break_point - 1],
                )
                print("guess and bnds", segment_guess, segment_bounds)
            hessian = None  # if bigauss_fit else gauss_hess

            results = [
                optimize.minimize(
                    fit_func,
                    segment_guess,
                    args,
                    method=routine,
                    bounds=segment_bounds,
                    options=opts,
                    jac=jacobian,
                    hess=hessian,
                    tol=1e-3,
                )
            ]
            while not results[-1].success and routines:
                routine = routines.pop(0)
                results.append(
                    optimize.minimize(
                        fit_func,
                        segment_guess,
                        args,
                        method=routine,
                        bounds=segment_bounds,
                        options=opts,
                        jac=jacobian,
                    )
                )
            if results[-1].success:
                res = results[-1]
            else:
                res = sorted(results, key=attrgetter("fun"))[0]
            n = len(segment_x)
            k = len(res.x)
            bic = n * np.log(res.fun / n) + k * np.log(n)
            res.bic = bic

            for index, value in enumerate(res.x[2::step_size]):
                if value < min_spacing:
                    res.x[2 + index * step_size] = min_spacing
            if bigauss_fit:
                for index, value in enumerate(res.x[3::step_size]):
                    if value < min_spacing:
                        res.x[3 + index * step_size] = min_spacing
            # does this data contain our rt peak?
            res._contains_rt = False
            if rt_peak != 0:
                for i in range(1, res.x.size, step_size):
                    mean = res.x[i]
                    lstd = res.x[i + 1]
                    if bigauss_fit:
                        rstd = res.x[i + 2]
                    else:
                        rstd = lstd
                    if (
                        rt_peak is not None
                        and mean - lstd * 2 < rt_peak < mean + rstd * 2
                    ):
                        res._contains_rt = True

            if not micro:
                # Rescale our data back
                # Amplitude
                res.x[::step_size] *= segment_max
                if fit_baseline:
                    # Slope
                    res.x[step_size - 2 :: step_size] *= segment_max
                    # Intercept
                    res.x[step_size - 1 :: step_size] *= segment_max

            # TODO: Evaluate the F-test based method
            # if best_res:
            #     cmodel_ssq = best_res.fun
            #     new_model_ssq = res.fun
            #     df = len(xdata)-len(res.x)
            #     f_ratio = (cmodel_ssq-new_model_ssq)/(new_model_ssq/df)
            #     res.p = 1-stats.f.cdf(f_ratio, 1, df)
            #     bic = res.p
            #
            # if not best_res or res.p < best_res.p:
            #     best_res = res
            #     best_fit = np.copy(res.x)
            #     best_rss = res.fun

            fitted_segments[(peak_width, chunk_index)].append((bic, res))

    # Figure out the best set of fits
    segment_order = sorted(fitted_segments.keys(), key=itemgetter(0, 1))
    best_fits = {
        peak_width: {"fit": [], "residual": 0,}
        for (peak_width, chunk_index) in segment_order
    }
    for key in segment_order:
        peak_width = key[0]
        fits = fitted_segments[key]
        lowest_bic = np.inf
        best_segment_res = 0
        best_segment_fit = None
        for bic, res in fits:
            if bic < lowest_bic or (
                getattr(best_segment_res, "_contains_rt", False) is not True
                and res._contains_rt is True
            ):
                if debug:
                    sys.stderr.write("{} < {}\n".format(bic, lowest_bic))
                if (
                    res._contains_rt is False
                    and best_segment_res != 0
                    and best_segment_res._contains_rt is True
                ):
                    continue
                if debug:
                    print("NEW BEST!", res, "old was", best_segment_res)
                best_segment_fit = np.copy(res.x)
                best_segment_res = res
                lowest_bic = bic
            if debug:
                sys.stderr.write("{} - best: {}\n".format(res, best_segment_fit))
        else:
            if best_segment_fit is None:
                return np.array([]), np.inf
            best_fits[peak_width]["fit"] += best_segment_fit.tolist()
            best_fits[peak_width]["residual"] += lowest_bic
            best_fits[peak_width][
                "contains_rt"
            ] = not best_segment_res._contains_rt  # this is so it sorts lower

    best_fit = sorted(
        (
            (
                best_fits[key[0]]["contains_rt"],
                best_fits[key[0]]["residual"],
                best_fits[key[0]]["fit"],
            )
            for key in segment_order
        ),
        key=itemgetter(0, 1),
    )

    if not best_fit:
        return (np.array([]), np.inf)

    best_fit = np.array(best_fit[0][2])

    # If the user only wants a certain number of peaks, enforce that now
    if max_peaks != -1:
        # pick the top n peaks for max_peaks
        if rt_peak:
            means = best_fit[1::step_size]
            # If the user specified a retention time as a guide, select the n peaks closest
            peak_indices = np.argsort(np.abs(means - rt_peak))[:max_peaks]
        else:
            # Return the top n highest peaks
            amplitudes = best_fit[0::step_size]
            peak_indices = sorted(np.argsort(amplitudes)[::-1][:max_peaks])
        best_fit = np.hstack(
            [best_fit[i * step_size : (i + 1) * step_size] for i in peak_indices]
        )

    peak_func = bigauss_ndim if bigauss_fit else gauss_ndim
    # Get rid of peaks with low r^2
    if not micro and r2_cutoff is not None:
        final_fit = np.array([])
        for peak_index in range(0, len(best_fit), step_size):

            peak_info = best_fit[peak_index : peak_index + step_size]
            amplitude, mean, std = peak_info[:3]
            left = mean - 2 * std
            right = mean + 2 * peak_info[3] if bigauss_fit else mean + 2 * std

            # Establish a goodness of fit using the coefficient of determination (the r^2) value for each peak.
            # Because the input data can have multiple peaks, we calculate a r^2 that considers the variance around this peak.
            curve_indices = (xdata >= left) & (xdata <= right)
            fitted_data = ydata[curve_indices]
            fitted_x = xdata[curve_indices]
            for other_peak_index in range(0, len(best_fit), step_size):
                if other_peak_index == peak_index:
                    continue
                fitted_data -= peak_func(
                    fitted_x, best_fit[other_peak_index : other_peak_index + step_size]
                )
            ss_tot = np.sum((fitted_data - np.mean(fitted_data)) ** 2)
            explained_data = peak_func(fitted_x, peak_info)
            if fit_baseline:
                explained_data += fitted_x * peak_info[-2] + peak_info[-1]
            ss_res = np.sum((fitted_data - explained_data) ** 2)
            coeff_det = 1 - (ss_res / ss_tot)
            if coeff_det >= r2_cutoff:
                final_fit = np.hstack((final_fit, peak_info))

        best_fit = final_fit

    residual = sum((ydata - peak_func(xdata, best_fit)) ** 2)

    if rescale:  # and not fit_baseline:
        best_fit[::step_size] *= original_max
        if fit_baseline:
            # Slope
            best_fit[step_size - 2 :: step_size] *= original_max
            # Intercept
            best_fit[step_size - 1 :: step_size] *= original_max

    return best_fit, residual


def findMicro(
    xdata,
    ydata,
    pos,
    ppm=None,
    start_mz=None,
    calc_start_mz=None,
    isotope=0,
    spacing=0,
    quant_method="integrate",
    fragment_scan=False,
    centroid=False,
    reporter_mode=False,
):
    """
    We want to find the boundaries of our isotopic clusters. Basically we search until our gradient
    changes, this assumes it's roughly gaussian and there is little interference
    """
    # find the edges within our tolerance
    tolerance = ppm
    offset = spacing * isotope
    df_empty_index = xdata[ydata == 0]
    if start_mz is None:
        start_mz = xdata[pos]
    fit = True
    if centroid:
        int_val = ydata[pos]
        left, right = pos - 1, pos + 1
        error = get_ppm(start_mz + offset, xdata[pos])
        fit = np.abs(error) < tolerance
        peak = [int_val, xdata[pos], 0]
    else:
        if df_empty_index.size == 0 or not (
            df_empty_index[0] < xdata[pos] < df_empty_index[-1]
        ):
            left = 0
            right = xdata.size
        else:
            right = np.searchsorted(df_empty_index, xdata[pos])
            left = right - 1
            left, right = (
                np.searchsorted(xdata, df_empty_index[left], side="left"),
                np.searchsorted(xdata, df_empty_index[right]),
            )
            right += 1
        new_x = xdata[left:right]
        new_y = ydata[left:right]
        if new_y.sum() == new_y.max():
            peak_mean = new_x[np.where(new_y > 0)][0]
            peaks = (new_y.max(), peak_mean, 0)
            sorted_peaks = [(peaks, get_ppm(start_mz + offset, peak_mean))]
        else:
            peaks, peak_residuals = findAllPeaks(
                new_x,
                new_y,
                min_dist=(new_x[1] - new_x[0]) * 2.0,
                peak_width_start=1,
                micro=True,
            )
            sorted_peaks = sorted(
                [
                    (peaks[i * 3 : (i + 1) * 3], get_ppm(start_mz + offset, v))
                    for i, v in enumerate(peaks[1::3])
                ],
                key=itemgetter(1),
            )
        if (fragment_scan is False or reporter_mode) and not within_tolerance(
            sorted_peaks, tolerance
        ):
            if calc_start_mz is not None:
                sorted_peaks2 = sorted(
                    [
                        (peaks[i * 3 : (i + 1) * 3], get_ppm(calc_start_mz + offset, v))
                        for i, v in enumerate(peaks[1::3])
                    ],
                    key=itemgetter(1),
                )
                if any(filter(lambda x: x[1] < tolerance, sorted_peaks2)):
                    sorted_peaks = sorted_peaks2
                else:
                    fit = False
            else:
                fit = False

        peak = np.array(sorted_peaks[0][0])
        # only go ahead with fitting if we have a stdev. Otherwise, set this to 0
        if peak[2] > 0:
            # peak[0] *= new_y.max()
            int_val = gauss_ndim(new_x, peak).sum()
        else:
            int_val = 0.0
        if not fit:
            pass
        error = sorted_peaks[0][1]
    ret_dict = {
        "int": int_val if fit or (fragment_scan is True and not reporter_mode) else 0.0,
        "int2": int_val,
        "bounds": (left, right),
        "params": peak,
        "error": error,
    }
    return ret_dict


def within_tolerance(arr, tolerance):
    # arr is a list of tuples with the [1] index for each tuple being the ppm error
    for i in arr:
        if i[1] < tolerance:
            return 1
    return 0


def targeted_search(
    merged_x,
    merged_y,
    x_value,
    attempts=4,
    max_peak_distance=1.0,
    peak_finding_kwargs=None,
):
    rt_attempts = 0
    fitting_y = np.copy(merged_y)
    find_peaks_kwargs = {
        "filter": False,
        "bigauss_fit": True,
        "rt_peak": x_value,
    }
    peak_finding_kwargs = peak_finding_kwargs or {}
    if peak_finding_kwargs:
        assert isinstance(
            peak_finding_kwargs, dict
        ), "peak_finding_kwargs must be a dictionary"
        find_peaks_kwargs.update(peak_finding_kwargs)
    debug = peak_finding_kwargs.get("debug")
    found_rt = False
    stepsize = 3
    if find_peaks_kwargs.get("bigauss_fit"):
        stepsize += 1
    if find_peaks_kwargs.get("fit_baseline"):
        stepsize += 2
    while rt_attempts < attempts and not found_rt:
        logger.debug("MERGED PEAK FINDING ATTEMPT %s", rt_attempts)
        res, residual = findAllPeaks(merged_x, fitting_y, **find_peaks_kwargs)
        if not res.any():
            return (None, np.inf)
        rt_peak = bigauss_ndim(np.array([x_value]), res)[0]
        # we don't do this routine for cases where there are > 5
        found_rt = sum(fitting_y > 0) <= 5 or rt_peak > 0.05
        if not found_rt and rt_peak < 0.05:
            # get the closest peak
            nearest_peak = sorted(
                [(i, np.abs(x_value - i)) for i in res[1::stepsize]], key=itemgetter(1)
            )[0][0]
            # this is tailored to mass spectrometry elution profiles at the moment, and only evaluates for situtations where the rt and peak
            # are no further than a minute apart.
            if np.abs(nearest_peak - x_value) < max_peak_distance:
                rt_index = find_nearest_index(merged_x, x_value)
                peak_index = find_nearest_index(merged_x, nearest_peak)
                if rt_index < 0:
                    rt_index = 0
                if peak_index == -1:
                    peak_index = len(fitting_y)
                if rt_index != peak_index:
                    grad_len = np.abs(peak_index - rt_index)
                    if grad_len < 4:
                        found_rt = True
                    else:
                        gradient = (
                            (np.gradient(fitting_y[rt_index:peak_index]) > 0)
                            if rt_index < peak_index
                            else (np.gradient(fitting_y[peak_index:rt_index]) < 0)
                        )
                        if sum(gradient) >= grad_len - 1:
                            found_rt = True
                else:
                    found_rt = True
        if not found_rt:
            if debug:
                print("cannot find rt for", x_value)
                print(merged_x, fitting_y, res, sum(fitting_y > 0))

            fitting_y -= bigauss_ndim(merged_x, res)
            fitting_y[fitting_y < 0] = 0
        rt_attempts += 1

    return (res, residual) if found_rt else (None, np.inf)



import os
import unittest
import pandas as pd
import subprocess

from . import mixins, utils


class ITraqTest(mixins.FileMixins, unittest.TestCase):
    def setUp(self):
        super(ITraqTest, self).setUp()
        self.labels = os.path.join(self.data_dir, "itraq_labels.tsv")
        self.output = os.path.join(self.out_dir, "itraq_test")

    @utils.timer
    def test_itraq_processing(self):
        com = [
            self.executable,
            "--scan-file",
            self.itraq_mzml,
            "-o",
            self.output,
            "--precursor-ppm",
            "200",
            "--isobaric-tags",
            "--label-scheme",
            self.labels,
        ]
        subprocess.call(com)
        data = pd.read_csv(self.output, sep="\t")
        for i, j in zip(
            [int(i) for i in data["114 Intensity"].values.tolist()],
            [1215, 9201, 1218, 83983, 10266, 2995, 7160],
        ):
            delta = j * 0.1
            if delta != 0 and delta < 300:
                delta = 300
            self.assertAlmostEqual(i, j, delta=delta)
        for i, j in zip(
            [int(i) for i in data["115 Intensity"].values.tolist()],
            [1428, 38772, 946, 4041, 12032, 4109, 8421],
        ):
            delta = j * 0.1
            if delta != 0 and delta < 300:
                delta = 300
            self.assertAlmostEqual(i, j, delta=delta)
        for i, j in zip(
            [int(i) for i in data["116 Intensity"].values.tolist()],
            [1031, 15314, 0, 94729, 11381, 3350, 8577],
        ):
            delta = j * 0.1
            if delta != 0 and delta < 300:
                delta = 300
            self.assertAlmostEqual(i, j, delta=delta)
        for i, j in zip(
            [int(i) for i in data["117 Intensity"].values.tolist()],
            [0, 42004, 1032, 164251, 14290, 2557, 8969],
        ):
            delta = j * 0.1
            if delta != 0 and delta < 300:
                delta = 300
            self.assertAlmostEqual(i, j, delta=delta)


if __name__ == "__main__":
    unittest.main()



from __future__ import print_function
import os

import numpy
from setuptools import (
    Extension,
    setup,
    find_packages,
)


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


extensions = [Extension(name="pyquant.cpeaks", sources=["pyquant/cpeaks.pyx"])]


setup(
    name="pyquant",
    version="0.3.2rc2",
    packages=find_packages(),
    scripts=["scripts/pyQuant"],
    entry_points={"console_scripts": ["pyQuant = pyquant.command_line:run_pyquant",]},
    install_requires=[
        "cython",
        "numpy",
        "scipy >= 0.18.*",
        "patsy",
        "pythomics >= 0.3.41",
        "pandas >= 0.24.0",
        "lxml",
        "scikit-learn",
        "simplejson",
    ],
    include_package_data=True,
    description="A framework for the analysis of quantitative mass spectrometry data",
    url="http://www.github.com/chris7/pyquant",
    author="Chris Mitchell <chris.mit7@gmail.com>",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    setup_requires=["cython", "numpy"],
    ext_modules=extensions,
    include_dirs=[numpy.get_include()],
)
